/*  core_shntool.c - functions to handle mode verification and execution
 *  Copyright (C) 2000-2001  Jason Jordan (shnutils@freeshell.org)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include "shntool.h"
#include "misc.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

char *progname,
     *progmode = NULL,
     fullprogname[PROGNAME_SIZE],
     *author = AUTHOR,
     *version = RELEASE,
     *copyright = COPYRIGHT,
     *url = URL;

int aliased = 0;

void show_supported_modules()
{
  int i;

  printf("Supported modes:\n");
  printf("\n");
  for (i=0;modes[i];i++)
    printf("  %" MAX_MODULE_NAME_LENGTH "s  (%s)\n",modes[i]->name,modes[i]->description);
  printf("\n");
  printf("Supported file formats:\n");
  for (i=0;formats[i];i++) {
    printf("\n");
    printf("  %" MAX_MODULE_NAME_LENGTH "s  (%s)\n",formats[i]->name,formats[i]->description);
    printf("    %" MAX_MODULE_NAME_LENGTH "s+ supports ","");
    if (0 != formats[i]->supports_input && 0 != formats[i]->supports_output)
      printf("input and output\n");
    else {
      if (0 != formats[i]->supports_input)
        printf("input only\n");
      else
        printf("output only\n");
    }
    if (0 != formats[i]->is_compressed)
      printf("    %" MAX_MODULE_NAME_LENGTH "s+ is a compression format\n","");
    if (0 != formats[i]->needs_external_program)
      printf("    %" MAX_MODULE_NAME_LENGTH "s+ relies on '%s' to function\n","",formats[i]->external_program);
  }
}

void show_usage()
{
  int i;

  printf("Usage: %s mode ...\n",progname);
  printf("       %s [OPTIONS]\n",progname);
  printf("\n");
  printf("  mode  A mode from the following list:\n");
  printf("\n");
  for (i=0;modes[i];i++)
    printf("      %" MAX_MODULE_NAME_LENGTH "s  (%s)\n",modes[i]->name,modes[i]->description);
  printf("\n");
  printf("        For help with any of the above modes, type:\n");
  printf("\n");
  printf("          %s mode -h\n",progname);
  printf("\n");
  printf("Options:\n");
  printf("\n");
  printf("  -m    shows detailed module information\n");
  printf("  -v    shows version information\n");
  printf("  -h    shows this help screen\n");
  printf("\n");

  exit(1);
}

void module_sanity_check()
{
  int i;

  if (modes[0]) {
    for (i=0;modes[i];i++) {
      if (NULL == modes[i]->name) {
        error("fatal: found a mode module without a name");
        exit(1);
      }
      if (NULL == modes[i]->description) {
        error("fatal: mode module '%s' has no description",modes[i]->name);
        exit(1);
      }
      if (NULL == modes[i]->run_main) {
        error("fatal: mode module '%s' does not provide run_main()",modes[i]->name);
        exit(1);
      }
    }

    for (i=1;modes[i];i++)
      if (0 == strcmp(modes[0]->name,modes[i]->name)) {
        error("fatal: found modes with identical name: %s",modes[0]->name);
        exit(1);
      }
  }
  else {
    error("fatal: no mode modules found");
    exit(1);
  }

  if (formats[0]) {
    for (i=0;formats[i];i++) {
      if (NULL == formats[i]->name) {
        error("fatal: found a format module without a name");
        exit(1);
      }
      if (NULL == formats[i]->description) {
        error("fatal: format module '%s' has no description",formats[i]->name);
        exit(1);
      }
      if (NULL == formats[i]->extension) {
        error("fatal: format module '%s' has no extension",formats[i]->name);
        exit(1);
      }
      if (0 != formats[i]->needs_external_program && NULL == formats[i]->external_program) {
        error("fatal: format module '%s' needs an external program to function, but none is specified",formats[i]->name);
        exit(1);
      }
      if (NULL == formats[i]->is_our_file) {
        error("fatal: format module '%s' does not provide is_our_file()",formats[i]->name);
        exit(1);
      }
      if (0 != formats[i]->supports_input && NULL == formats[i]->open_input) {
        error("fatal: format module '%s' supports input, but does not provide open_input()",formats[i]->name);
        exit(1);
      }
      if (0 != formats[i]->supports_output && NULL == formats[i]->open_output) {
        error("fatal: format module '%s' supports output, but does not provide open_output()",formats[i]->name);
        exit(1);
      }
      if (0 == formats[i]->supports_input && 0 == formats[i]->supports_output) {
        error("fatal: format module '%s' doesn't support input or output",formats[i]->name);
        exit(1);
      }
    }

    for (i=1;formats[i];i++)
      if (0 == strcmp(formats[0]->name,formats[i]->name)) {
        error("fatal: found formats with identical name: %s",formats[0]->name);
        exit(1);
      }
  }
  else {
    error("fatal: no file formats modules found");
    exit(1);
  }
}

void parse_main(int argc,char **argv)
{
  int i;

  /* look for a module alias matching progname */

  for (i=0;modes[i];i++) {
    if (NULL != modes[i]->alias && 0 == strcmp(progname,modes[i]->alias)) {
      aliased = 1;
      progmode = modes[i]->name;
      modes[i]->run_main(argc,argv,1);
      return;
    }
  }

  /* failing that, look for a module name matching argv[1] */

  if (argc < 2) {
    error("missing arguments");
    type_h_for_help();
  }

  for (i=0;modes[i];i++) {
    if (0 == strcmp(argv[1],modes[i]->name)) {
      /* found mode - now run it and quit */
      progmode = modes[i]->name;
      strcat(fullprogname," ");
      strcat(fullprogname,progmode);
      modes[i]->run_main(argc,argv,2);
      return;
    }
  }

  if (argc > 1 && 0 == strcmp(argv[1],"-h"))
    show_usage();
  else if (argc > 1 && 0 == strcmp(argv[1],"-v"))
    internal_version();
  else if (argc > 1 && 0 == strcmp(argv[1],"-m"))
    show_supported_modules();
  else {
    /* didn't find any matching modes */
    error("invalid mode or option: %s",argv[1]);
    type_h_for_help();
  }
}

int main(int argc,char **argv)
{
  /* set up globals */
  progname = (0 == strrchr(argv[0],'/')) ? argv[0] : strrchr(argv[0],'/') + 1;
  strcpy(fullprogname,progname);

  /* sanity check for modules */
  module_sanity_check();

  /* ignore dying children */
  signal(SIGCHLD,SIG_IGN);

  /* parse command line */
  parse_main(argc,argv);

  return 0;
}
