/*  mode_cmp.c - cmp mode module
 *  Copyright (C) 2000-2001  Jason Jordan (shnutils@freeshell.org)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "shntool.h"
#include "fileio.h"
#include "wave.h"
#include "misc.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

void cmp_main(int,char **,int);

mode_module mode_cmp = {
  "cmp",
  "shncmp",
  "compares PCM WAVE data in two files",
  cmp_main
};

void show_usage_cmp()
{
  printf("Usage: %s [OPTIONS] [file1 file2]\n",fullprogname);
  printf("\n");
  printf("Options:\n");
  printf("\n");
  printf("  --     indicates that everything following it is a file name\n");
  printf("\n");
  printf("  -v     shows version information\n");
  printf("  -h     shows this help screen\n");
  printf("\n");
  printf("  If no file names are given, then file names are read from standard input.\n");
  printf("\n");
  printf("  ------------\n");
  printf("  Sample uses:\n");
  printf("  ------------\n");
  printf("\n");
  printf("  %% %s filename1.wav filename2.shn\n",fullprogname);
  printf("\n");

  exit(1);
}

void parse_cmp(int argc,char **argv,int start,int *first_arg)
{
  while (start < argc && argv[start][0] == '-') {
    if (argc > start && 0 == strcmp(argv[start],"--")) {
      start++;
      break;
    }
    else if (argc > start && 0 == strcmp(argv[start],"-v"))
      internal_version();
    else if (argc > start && 0 == strcmp(argv[start],"-h"))
      show_usage_cmp();
    else {
      error("unknown argument: %s",argv[start]);
      type_h_for_help();
    }
    start++;
  }

  if (start != argc && start != argc - 2) {
    error("need exactly two files to process");
    type_h_for_help();
  }

  *first_arg = start;
}

void check_headers(wave_info *info1,wave_info *info2)
{
  if (info1->wave_format != info2->wave_format) {
    error("WAVE format differs between these files");
    exit(1);
  }

  if (info1->channels != info2->channels) {
    error("number of channels differs between these files");
    exit(1);
  }

  if (info1->samples_per_sec != info2->samples_per_sec) {
    error("samples per second differs between these files");
    exit(1);
  }

  if (info1->avg_bytes_per_sec != info2->avg_bytes_per_sec) {
    error("average bytes per second differs between these files");
    exit(1);
  }

  if (info1->bits_per_sample != info2->bits_per_sample) {
    error("bits per sample differs between these files");
    exit(1);
  }

  if (info1->block_align != info2->block_align)
    printf("warning: block align differs between these files\n");

  if (info1->data_size != info2->data_size) {
    printf("warning: data size differs between these files - WAVE data will\n");
    printf("         only be compared up to the size of the smaller file\n");
  }
}

int my_memcmp(unsigned char *str1,unsigned char *str2,int len)
{
  int i;

  for (i=0;i<len;i++) {
    if (str1[i] > str2[i] || str1[i] < str2[i])
      return i+1;
  }

  return 0;
}

void cmp_files(wave_info *info1,wave_info *info2)
{
  unsigned char *header,
                buf1[XFER_SIZE],
                buf2[XFER_SIZE];
  wlong bytes_to_check,
        bytes_checked = 0,
        bytes;
  int offset;

  check_headers(info1,info2);

  if (NULL == (info1->input = info1->input_format->open_input(info1->filename))) {
    error("could not reopen file '%s' for input",info1->filename);
    exit(1);
  }

  if (NULL == (info2->input = info2->input_format->open_input(info2->filename))) {
    error("could not reopen file '%s' for input",info2->filename);
    exit(1);
  }

  if (NULL == (header = malloc(max(info1->header_size,info2->header_size) * sizeof(unsigned char)))) {
    error("could not allocate %d bytes for WAVE header",max(info1->header_size,info2->header_size));
    exit(1);
  }

  if (read_n_bytes(info1->input,header,info1->header_size) != info1->header_size) {
    error("error while discarding %d-byte header for file '%s'",info1->header_size,info1->filename);
    exit(1);
  }

  if (read_n_bytes(info2->input,header,info2->header_size) != info2->header_size) {
    error("error while discarding %d-byte header for file '%s'",info2->header_size,info2->filename);
    exit(1);
  }

  printf("comparing WAVE data in '%s' and '%s'\n",info1->filename,info2->filename);

  bytes_to_check = min(info1->data_size,info2->data_size);

  while (bytes_to_check > 0) {
    bytes = min(bytes_to_check,XFER_SIZE);
    if (read_n_bytes(info1->input,buf1,(int)bytes) != (int)bytes) {
      error("error while reading %d bytes from '%s'",(int)bytes,info1->filename);
      exit(1);
    }
    if (read_n_bytes(info2->input,buf2,(int)bytes) != (int)bytes) {
      error("error while reading %d bytes from '%s'",(int)bytes,info2->filename);
      exit(1);
    }
    if (0 != (offset = my_memcmp(buf1,buf2,bytes))) {
      error("WAVE data differs at byte offset %lu",bytes_checked+(wlong)offset);
      exit(1);
    }
    bytes_to_check -= bytes;
    bytes_checked += bytes;
  }

  fclose(info1->input);
  fclose(info2->input);

  printf("contents of these files are identical");
  if (info1->data_size != info2->data_size)
    printf(" (up to the first %lu bytes of WAVE data)",min(info1->data_size,info2->data_size));
  printf("\n");

  free(header);
}

void process_files_cmp(char *filename1,char *filename2)
{
  wave_info *info1,*info2;

  if (NULL == (info1 = new_wave_info(filename1)))
    return;

  if (NULL == (info2 = new_wave_info(filename2)))
    return;

  cmp_files(info1,info2);

  free(info1);
  free(info2);
}

void process_cmp(int argc,char **argv,int start)
{
  char filename1[FILENAME_SIZE],
       filename2[FILENAME_SIZE];

  if (argc < start + 1) {
    /* no filenames were given, so we're reading two file names from standard input. */
    fgets(filename1,FILENAME_SIZE-1,stdin);
    if (0 == feof(stdin)) {
      fgets(filename2,FILENAME_SIZE-1,stdin);
      if (0 == feof(stdin)) {
        filename1[strlen(filename1)-1] = '\0';
        filename2[strlen(filename2)-1] = '\0';
        process_files_cmp(filename1,filename2);
      }
      else
        error("need two file names to process");
    }
  }
  else
    process_files_cmp(argv[start],argv[start+1]);
}

void cmp_main(int argc,char **argv,int start)
{
  int first_arg;

  parse_cmp(argc,argv,start,&first_arg);

  process_cmp(argc,argv,first_arg);
}
