/*  mode_fix.c - fix mode module
 *  Copyright (C) 2000-2001  Jason Jordan (shnutils@freeshell.org)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "shntool.h"
#include "fileio.h"
#include "wave.h"
#include "misc.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define POSTFIX_FIXED "fixed"

typedef enum {
  SHIFT_UNKNOWN,
  SHIFT_BACKWARD,
  SHIFT_FORWARD,
  SHIFT_ROUND
} shifts;

int fix_order = 0,
    fix_pad = 1,
    fix_pad_bytes = 0,
    fix_preview = 0,
    fix_skip = 1,
    fix_warned = 0,
    fix_output_dir_flag = 0,
    desired_shift = SHIFT_UNKNOWN,
    fix_numfiles;

char *fix_output_directory = ".";

format_module *fix_op = NULL;

wave_info **fix_files;

void fix_main(int,char **,int);

mode_module mode_fix = {
  "fix",
  "shnfix",
  "fixes sector-boundary problems with CD-quality PCM WAVE data",
  fix_main
};

void show_usage_fix()
{
  int i,j=0;

  printf("Usage: %s [OPTIONS] file1 [file2 ...]\n",fullprogname);
  printf("\n");
  printf("Options:\n");
  printf("\n");
  printf("  -o fmt   specifies the output file format, where fmt is one of the following:\n");
  printf("\n");
  for (i=0;formats[i];i++)
    if (0 != formats[i]->supports_output)
      printf("             %" MAX_MODULE_NAME_LENGTH "s  (%s) %s\n",formats[i]->name,formats[i]->description,(0 == (j++))?"[DEFAULT]":"");
  printf("\n");
  printf("  -s type  specifies the shift type, where type is one of the following:\n");
  printf("\n");
  printf("             b  shifts track breaks backward to the previous multiple of the\n");
  printf("                block size (%d bytes) [DEFAULT]\n",CD_BLOCK_SIZE);
  printf("             f  shifts track breaks forward to the next multiple of the block\n");
  printf("                size (%d bytes)\n",CD_BLOCK_SIZE);
  printf("             r  rounds track breaks to the nearest multiple of the block size\n");
  printf("                (%d bytes)\n",CD_BLOCK_SIZE);
  printf("\n");
  printf("  -d dir   specifies alternate output directory (default is current directory)\n");
  printf("\n");
  printf("  -nopad   specifies that zero-padding should not be done (default is to pad)\n");
  printf("\n");
  printf("  -noskip  don't skip first N files that won't be changed from a WAVE data\n");
  printf("           perspective (default is to skip them to avoid unnecessary work)\n");
  printf("\n");
  printf("  -order   allows you to edit the order of the files before processing them\n");
  printf("           (useful for certain filenames, where *.shn might give d1t1.shn,\n");
  printf("            d1t10.shn, d1t11.shn, d1t2.shn, d1t3.shn, ...)\n");
  printf("\n");
  printf("  -p       preview what changes would be made without actually making them\n");
  printf("\n");
  printf("  --       indicates that everything following it is a file name\n");
  printf("\n");
  printf("  -v       shows version information\n");
  printf("  -h       shows this help screen\n");
  printf("\n");
  printf("  ------------\n");
  printf("  Sample uses:\n");
  printf("  ------------\n");
  printf("\n");
  printf("    %% %s -o wav -order *.shn\n",fullprogname);
  printf("\n");
  printf("    %% %s -s r -o shn -nopad *.wav\n",fullprogname);
  printf("\n");

  exit(1);
}

void parse_fix(int argc,char **argv,int start,int *first_arg)
{
  int i,found_output = -1;

  for (i=0;formats[i];i++)
    if (0 != formats[i]->supports_output) {
      found_output = i;
      break;
    }

  if (-1 == found_output) {
    error("no output formats found, cannot continue");
    exit(1);
  }

  while (start < argc && argv[start][0] == '-') {
    if (argc > start && 0 == strcmp(argv[start],"--")) {
      start++;
      break;
    }
    else if (argc > start && 0 == strcmp(argv[start],"-v"))
      internal_version();
    else if (argc > start && 0 == strcmp(argv[start],"-h"))
      show_usage_fix();
    else if (argc > start && 0 == strcmp(argv[start],"-order"))
      fix_order = 1;
    else if (argc > start && 0 == strcmp(argv[start],"-nopad"))
      fix_pad = 0;
    else if (argc > start && 0 == strcmp(argv[start],"-noskip"))
      fix_skip = 0;
    else if (argc > start && 0 == strcmp(argv[start],"-p"))
      fix_preview = 1;
    else if (argc > start && 0 == strcmp(argv[start],"-d")) {
      if (1 == fix_output_dir_flag) {
        error("too many output directories specified");
        type_h_for_help();
      }
      start++;
      fix_output_dir_flag = 1;
      if (argc <= start) {
        error("missing output directory");
        type_h_for_help();
      }
      fix_output_directory = argv[start];
    }
    else if (argc > start && 0 == strcmp(argv[start],"-s")) {
      start++;
      if (argc <= start) {
        error("missing shift type");
        type_h_for_help();
      }
      if (0 == strcmp(argv[start],"b")) {
        if (desired_shift != SHIFT_UNKNOWN && desired_shift != SHIFT_BACKWARD) {
          error("too many shift types specified");
          type_h_for_help();
        }
        desired_shift = SHIFT_BACKWARD;
      }
      else if (0 == strcmp(argv[start],"f")) {
        if (desired_shift != SHIFT_UNKNOWN && desired_shift != SHIFT_FORWARD) {
          error("too many shift types specified");
          type_h_for_help();
        }
        desired_shift = SHIFT_FORWARD;
      }
      else if (0 == strcmp(argv[start],"r")) {
        if (desired_shift != SHIFT_UNKNOWN && desired_shift != SHIFT_ROUND) {
          error("too many shift types specified");
          type_h_for_help();
        }
        desired_shift = SHIFT_ROUND;
      }
      else {
        error("unknown shift type: %s",argv[start]);
        type_h_for_help();
      }
    }
    else if (argc > start && 0 == strcmp(argv[start],"-o")) {
      if (NULL != fix_op) {
        error("too many output file formats specified");
        type_h_for_help();
      }
      start++;
      if (argc <= start) {
        error("missing output file format");
        type_h_for_help();
      }
      for (i=0;formats[i];i++) {
        if (0 == strcmp(argv[start],formats[i]->name)) {
          if (0 == formats[i]->supports_output) {
            error("file format '%s' does not support output",formats[i]->name);
            type_h_for_help();
          }
          fix_op = formats[i];
          break;
        }
      }
      if (NULL == fix_op) {
        error("invalid output file format: %s",argv[start]);
        type_h_for_help();
      }
    }
    else {
      error("unknown argument: %s",argv[start]);
      type_h_for_help();
    }
    start++;
  }

  if (start >= argc) {
    error("need two or more files to process");
    type_h_for_help();
  }

  if (NULL == fix_op) {
    fix_op = formats[found_output];
    printf("warning: no output format specified - defaulting to %s\n",fix_op->name);
    fix_warned = 1;
  }

  if (SHIFT_UNKNOWN == desired_shift) {
    desired_shift = SHIFT_BACKWARD;
    printf("warning: no shift direction specified - assuming backward shift\n");
    fix_warned = 1;
  }

  *first_arg = start;
}

void calculate_breaks_backward()
{
  int i,remainder = 0;
  wlong tmp,begin_before = 0,begin_after = 0;

  for (i=0;i<fix_numfiles-1;i++) {
    fix_files[i]->beginning_byte = begin_before;
    fix_files[i]->new_beginning_byte = begin_after;

    tmp = fix_files[i]->data_size + remainder;
    remainder = tmp % CD_BLOCK_SIZE;
    fix_files[i]->new_data_size = tmp - remainder;

    begin_before += fix_files[i]->data_size;
    begin_after += fix_files[i]->new_data_size;
  }
  fix_files[fix_numfiles-1]->new_data_size = fix_files[fix_numfiles-1]->data_size + remainder;
  fix_files[fix_numfiles-1]->beginning_byte = begin_before;
  fix_files[fix_numfiles-1]->new_beginning_byte = begin_after;
}

void calculate_breaks_forward()
{
  int i,used = 0,remainder;
  wlong tmp,begin_before = 0,begin_after = 0;

  for (i=0;i<fix_numfiles-1;i++) {
    fix_files[i]->beginning_byte = begin_before;
    fix_files[i]->new_beginning_byte = begin_after;

    tmp = fix_files[i]->data_size - used;
    remainder = tmp % CD_BLOCK_SIZE;
    if (remainder) {
      used = CD_BLOCK_SIZE - remainder;
      fix_files[i]->new_data_size = tmp + used;
    }
    else {
      used = 0;
      fix_files[i]->new_data_size = tmp;
    }

    begin_before += fix_files[i]->data_size;
    begin_after += fix_files[i]->new_data_size;
  }
  fix_files[fix_numfiles-1]->new_data_size = fix_files[fix_numfiles-1]->data_size - used;
  fix_files[fix_numfiles-1]->beginning_byte = begin_before;
  fix_files[fix_numfiles-1]->new_beginning_byte = begin_after;
}

void calculate_breaks_round()
{
  int i,give_or_take = 0;
  wlong tmp,how_much,begin_before = 0,begin_after = 0;

  for (i=0;i<fix_numfiles-1;i++) {
    fix_files[i]->beginning_byte = begin_before;
    fix_files[i]->new_beginning_byte = begin_after;

    tmp = fix_files[i]->data_size + give_or_take;
    how_much = (((tmp + CD_BLOCK_SIZE/2) / CD_BLOCK_SIZE) * CD_BLOCK_SIZE);
    give_or_take = tmp - how_much;
    fix_files[i]->new_data_size = how_much;

    begin_before += fix_files[i]->data_size;
    begin_after += fix_files[i]->new_data_size;
  }
  fix_files[fix_numfiles-1]->new_data_size = fix_files[fix_numfiles-1]->data_size + give_or_take;
  fix_files[fix_numfiles-1]->beginning_byte = begin_before;
  fix_files[fix_numfiles-1]->new_beginning_byte = begin_after;
}

void calculation_sanity_check()
{
  int i;
  wlong old_total = 0,new_total = 0;

  for (i=0;i<fix_numfiles;i++) {
    old_total += fix_files[i]->data_size;
    new_total += fix_files[i]->new_data_size;
  }

  if (old_total != new_total) {
    error("total WAVE data size differs from newly calculated total -\n"
          "please file a bug report with the following data:");
    fprintf(stderr,"\nshift type: %s\n\n",(SHIFT_BACKWARD == desired_shift)?"backward":
                                         ((SHIFT_FORWARD == desired_shift)?"forward":
                                         ((SHIFT_ROUND == desired_shift)?"round":"unknown")));
    for (i=0;i<fix_numfiles;i++) {
      fprintf(stderr,"file %2d:  data size = %10lu, new data size = %10lu\n",i+1,fix_files[i]->data_size,fix_files[i]->new_data_size);
    }
    fprintf(stderr,"\ntotals :  data size = %10lu, new data size = %10lu\n",old_total,new_total);
    exit(1);
  }
}

void make_outfilename_shift(char *infilename,char *outfilename)
/* creates an output file name for shifted data, used in fix mode */
{
  char *basename;

  if ((basename = strrchr(infilename,'/')))
    basename++;
  else
    basename = infilename;

  if (1 == fix_output_dir_flag)
#ifdef HAVE_SNPRINTF
    snprintf(outfilename,FILENAME_SIZE,"%s/%s",fix_output_directory,basename);
#else
    sprintf(outfilename,"%s/%s",fix_output_directory,basename);
#endif
  else
#ifdef HAVE_SNPRINTF
    snprintf(outfilename,FILENAME_SIZE,"%s",basename);
#else
    sprintf(outfilename,"%s",basename);
#endif

  if (filename_contains_a_dot(outfilename))
    *(strrchr(outfilename,'.')) = '\0';

  strcat(outfilename,"-");
  strcat(outfilename,POSTFIX_FIXED);
  strcat(outfilename,".");
  strcat(outfilename,fix_op->extension);
}

void show_potential_changes()
{
  int i;
  char outfilename[FILENAME_SIZE];

  printf("\n");
  printf("Preview of changes:\n");
  printf("-------------------\n");
  printf("\n");
  printf("Track breaks will be %s when necessary.\n\n",
         (SHIFT_BACKWARD == desired_shift)?"shifted backward":
         ((SHIFT_FORWARD == desired_shift)?"shifted forward":
                                           "rounded to the nearest sector boundary"));

  for (i=0;i<fix_numfiles;i++) {
    make_outfilename_shift(fix_files[i]->filename,outfilename);
    printf("%s --> %s\n",fix_files[i]->filename,outfilename);

    printf("  - beginning of track will ");
    if (fix_files[i]->beginning_byte == fix_files[i]->new_beginning_byte)
      printf("remain unchanged");
    else {
      printf("be moved ");
      if (fix_files[i]->beginning_byte > fix_files[i]->new_beginning_byte)
        printf("backward by %lu bytes",fix_files[i]->beginning_byte - fix_files[i]->new_beginning_byte);
      else
        printf("forward by %lu bytes",fix_files[i]->new_beginning_byte - fix_files[i]->beginning_byte);
    }
    printf("\n");

    printf("  - file size will ");
    if (fix_files[i]->data_size == fix_files[i]->new_data_size)
      printf("remain unchanged");
    else {
      if (fix_files[i]->data_size > fix_files[i]->new_data_size)
        printf("decrease by %lu bytes",fix_files[i]->data_size - fix_files[i]->new_data_size);
      else
        printf("increase by %lu bytes",fix_files[i]->new_data_size - fix_files[i]->data_size);
    }
    printf("\n\n");
  }

  if (1 == fix_pad) {
    printf("The last file '%s' would",outfilename);
    if (0 != fix_pad_bytes)
      printf(" be padded with %d zero-bytes.\n",fix_pad_bytes);
    else
      printf("n't need any padding.\n");
  }
  else {
    printf("The last file '%s' wouldn't be padded, ",outfilename);
    if (0 != fix_pad_bytes)
      printf("but would need %d bytes of padding.\n",fix_pad_bytes);
    else
      printf("and wouldn't need it.\n");
  }
}

void write_fixed_files()
{
  int i,cur_input,cur_output;
  unsigned long bytes_have,bytes_needed,bytes_to_xfer;
  char outfilename[FILENAME_SIZE];
  unsigned char *header;

  for (i=0;i<fix_numfiles;i++) {
    if (NULL == (fix_files[i]->input = fix_files[i]->input_format->open_input(fix_files[i]->filename))) {
      error("could not reopen file '%s' for input",fix_files[i]->filename);
      exit(1);
    }
    make_outfilename_shift(fix_files[i]->filename,outfilename);
    if (NULL == (fix_files[i]->output = fix_op->open_output(outfilename))) {
      error("could not open file '%s' for output",outfilename);
      exit(1);
    }
    if (NULL == (header = malloc(fix_files[i]->header_size * sizeof(unsigned char)))) {
      error("could not allocate %d bytes for WAVE header",fix_files[i]->header_size);
      exit(1);
    }

    if (read_n_bytes(fix_files[i]->input,header,fix_files[i]->header_size) != fix_files[i]->header_size) {
      error("error while reading %d-byte header",fix_files[i]->header_size);
      exit(1);
    }

    if (fix_numfiles - 1 == i && 1 == fix_pad)
      put_data_size(header,fix_files[i]->header_size,fix_files[i]->new_data_size+fix_pad_bytes);
    else
      put_data_size(header,fix_files[i]->header_size,fix_files[i]->new_data_size);

    if (write_n_bytes(fix_files[i]->output,header,fix_files[i]->header_size) != fix_files[i]->header_size) {
      error("error while writing %d-byte header",fix_files[i]->header_size);
      exit(1);
    }

    free(header);
  }

  cur_input = cur_output = 0;
  bytes_have = (unsigned long)fix_files[0]->data_size;
  bytes_needed = (unsigned long)fix_files[0]->new_data_size;

  make_outfilename_shift(fix_files[0]->filename,outfilename);
  printf("%s --> %s ... ",fix_files[0]->filename,outfilename);
  fflush(stdout);

  while (cur_input < fix_numfiles && cur_output < fix_numfiles) {
    bytes_to_xfer = min(bytes_have,bytes_needed);
    if (transfer_n_bytes(fix_files[cur_input]->input,fix_files[cur_output]->output,bytes_to_xfer) != bytes_to_xfer) {
      error("error while transferring %lu bytes of data",bytes_to_xfer);
      exit(1);
    }

    bytes_have -= bytes_to_xfer;
    bytes_needed -= bytes_to_xfer;

    if (0 == bytes_have) {
      printf("done.\n");
      fclose(fix_files[cur_input++]->input);
      if (cur_input < fix_numfiles) {
        bytes_have = (unsigned long)fix_files[cur_input]->data_size;
        make_outfilename_shift(fix_files[cur_input]->filename,outfilename);
        printf("%s --> %s ... ",fix_files[cur_input]->filename,outfilename);
        fflush(stdout);
      }
    }

    if (0 == bytes_needed) {
      if (fix_numfiles - 1 == cur_output) {
        if (1 == fix_pad) {
          if (0 != fix_pad_bytes) {
            pad_file(fix_files[cur_output]->output,fix_pad_bytes);
            printf("Padded '%s' with %d zero-bytes.\n",outfilename,fix_pad_bytes);
          }
          else
            printf("No padding needed for '%s'.\n",outfilename);
        }
        else {
          printf("File '%s' was not padded, ",outfilename);
          if (0 != fix_pad_bytes)
            printf("though it needs %d bytes of padding.\n",fix_pad_bytes);
          else
            printf("nor was it needed.\n");
        }
      }
      fclose(fix_files[cur_output++]->output);
      if (cur_output < fix_numfiles)
        bytes_needed = (unsigned long)fix_files[cur_output]->new_data_size;
    }

  }
}

void process_files_fix(int argc,char **argv,int start)
{
  int i,j,remainder,
      badfiles = 0,
      needs_fixing = 0,
      found_errors = 0;

  if (NULL == (fix_files = malloc((fix_numfiles + 1) * sizeof(wave_info *)))) {
    error("could not allocate memory for file info array");
    exit(1);
  }

  for (i=0;i<fix_numfiles;i++)
    if (NULL == (fix_files[i] = new_wave_info(argv[start+i])))
      badfiles++;

  fix_files[fix_numfiles] = NULL;

  if (badfiles) {
    error("could not open %d file%s, see above error messages",badfiles,(badfiles>1)?"s":"");
    exit(1);
  }

  /* validate files */
  for (i=0;i<fix_numfiles;i++) {
    if (PROB_NOT_CD(fix_files[i])) {
      error("file '%s' is not CD-quality",fix_files[i]->filename);
      found_errors = 1;
    }
    if (PROB_HDR_INCONSISTENT(fix_files[i])) {
      error("file '%s' has an inconsistent header",fix_files[i]->filename);
      found_errors = 1;
    }
    if (PROB_TRUNCATED(fix_files[i])) {
      error("file '%s' seems to be truncated",fix_files[i]->filename);
      found_errors = 1;
    }
    if (PROB_BAD_BOUND(fix_files[i]))
      needs_fixing = 1;
  }

  if (0 != found_errors) {
    error("could not fix files due to errors, see above");
    exit(1);
  }

  if (0 == needs_fixing) {
    error("everything seems fine, no need for fixing");
    exit(1);
  }

  if (1 == fix_order)
    alter_file_order(fix_files,fix_numfiles);

  i = 0;
  while (0 == PROB_BAD_BOUND(fix_files[i]))
    i++;

  if (1 == fix_skip) {
    if (i != 0) {
      printf("warning: skipping first %d file%s because %s would not be changed\n\n",i,(1==i)?"":"s",(1==i)?"it":"they");
      for (j=0;j<i;j++)
        free(fix_files[j]);
      for (j=i;j<fix_numfiles;j++) {
        fix_files[j-i] = fix_files[j];
        fix_files[j] = NULL;
      }
      fix_numfiles -= i;
    }
  }
  else
    if (1 == fix_warned)
      printf("\n");

  if (1 == fix_numfiles && 0 == fix_pad) {
    error("file '%s' (the only file to process) won't be padded, so there is nothing to do",fix_files[0]->filename);
    exit(1);
  }

  switch (desired_shift) {
    case SHIFT_BACKWARD:
      calculate_breaks_backward();
      break;
    case SHIFT_FORWARD:
      calculate_breaks_forward();
      break;
    case SHIFT_ROUND:
      calculate_breaks_round();
      break;
  }

  calculation_sanity_check();

  remainder = fix_files[fix_numfiles-1]->new_data_size % CD_BLOCK_SIZE;
  if (0 != remainder)
    fix_pad_bytes = CD_BLOCK_SIZE - remainder;

  if (1 == fix_preview)
    show_potential_changes();
  else
    write_fixed_files();

  for (i=0;i<fix_numfiles;i++)
    free(fix_files[i]);

  free(fix_files);
}

void fix_main(int argc,char **argv,int start)
{
  int first_arg;

  parse_fix(argc,argv,start,&first_arg);

  fix_numfiles = argc-first_arg;

  process_files_fix(argc,argv,first_arg);
}
