/*  mode_join.c - join mode module
 *  Copyright (C) 2000-2001  Jason Jordan (shnutils@freeshell.org)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "shntool.h"
#include "fileio.h"
#include "wave.h"
#include "misc.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define PREFIX_JOIN   "joined"

int join_order = 0,
    join_pad = 1,
    join_pad_bytes = 0,
    join_preview = 0,
    join_output_dir_flag = 0,
    join_numfiles;

char *join_output_directory = ".";

format_module *join_op = NULL;

wave_info **join_files;

void join_main(int,char **,int);

mode_module mode_join = {
  "join",
  "shnjoin",
  "joins CD-quality PCM WAVE data from multiple files into one",
  join_main
};

void show_usage_join()
{
  int i,j=0;

  printf("Usage: %s [OPTIONS] file1 file2 [file3 ...]\n",fullprogname);
  printf("\n");
  printf("Options:\n");
  printf("\n");
  printf("  -o fmt   specifies the output file format, where fmt is one of the following:\n");
  printf("\n");
  for (i=0;formats[i];i++)
    if (0 != formats[i]->supports_output)
      printf("             %" MAX_MODULE_NAME_LENGTH "s  (%s) %s\n",formats[i]->name,formats[i]->description,(0 == (j++))?"[DEFAULT]":"");
  printf("\n");
  printf("  -d dir   specifies alternate output directory (default is current directory)\n");
  printf("\n");
  printf("  -nopad   specifies that zero-padding should not be done (default is to pad)\n");
  printf("\n");
  printf("  -order   allows you to edit the order of the files before processing them\n");
  printf("           (useful for certain filenames, where *.shn might give d1t1.shn,\n");
  printf("            d1t10.shn, d1t11.shn, d1t2.shn, d1t3.shn, ...)\n");
  printf("\n");
  printf("  -p       preview what changes would be made without actually making them\n");
  printf("\n");
  printf("  --       indicates that everything following it is a file name\n");
  printf("\n");
  printf("  -v       shows version information\n");
  printf("  -h       shows this help screen\n");
  printf("\n");
  printf("  ------------\n");
  printf("  Sample uses:\n");
  printf("  ------------\n");
  printf("\n");
  printf("    %% %s -o wav -order *.shn\n",fullprogname);
  printf("\n");
  printf("    %% %s -o shn -nopad -p *.wav\n",fullprogname);
  printf("\n");

  exit(1);
}

void parse_join(int argc,char **argv,int start,int *first_arg)
{
  int i,found_output = -1;

  for (i=0;formats[i];i++)
    if (0 != formats[i]->supports_output) {
      found_output = i;
      break;
    }

  if (-1 == found_output) {
    error("no output formats found, cannot continue");
    exit(1);
  }

  while (start < argc && argv[start][0] == '-') {
    if (argc > start && 0 == strcmp(argv[start],"--")) {
      start++;
      break;
    }
    else if (argc > start && 0 == strcmp(argv[start],"-v"))
      internal_version();
    else if (argc > start && 0 == strcmp(argv[start],"-h"))
      show_usage_join();
    else if (argc > start && 0 == strcmp(argv[start],"-order"))
      join_order = 1;
    else if (argc > start && 0 == strcmp(argv[start],"-nopad"))
      join_pad = 0;
    else if (argc > start && 0 == strcmp(argv[start],"-p"))
      join_preview = 1;
    else if (argc > start && 0 == strcmp(argv[start],"-d")) {
      if (1 == join_output_dir_flag) {
        error("too many output directories specified");
        type_h_for_help();
      }
      start++;
      join_output_dir_flag = 1;
      if (argc <= start) {
        error("missing output directory");
        type_h_for_help();
      }
      join_output_directory = argv[start];
    }
    else if (argc > start && 0 == strcmp(argv[start],"-o")) {
      if (NULL != join_op) {
        error("too many output file formats specified");
        type_h_for_help();
      }
      start++;
      if (argc <= start) {
        error("missing output file format");
        type_h_for_help();
      }
      for (i=0;formats[i];i++) {
        if (0 == strcmp(argv[start],formats[i]->name)) {
          if (0 == formats[i]->supports_output) {
            error("file format '%s' does not support output",formats[i]->name);
            type_h_for_help();
          }
          join_op = formats[i];
          break;
        }
      }
      if (NULL == join_op) {
        error("invalid output file format: %s",argv[start]);
        type_h_for_help();
      }
    }
    else {
      error("unknown argument: %s",argv[start]);
      type_h_for_help();
    }
    start++;
  }

  if (start >= argc - 1) {
    error("need two or more files to process");
    type_h_for_help();
  }

  if (NULL == join_op) {
    join_op = formats[found_output];
    printf("warning: no output format specified - defaulting to %s\n\n",join_op->name);
  }

  *first_arg = start;
}

void do_join()
{
  int i,bytes_to_skip,bytes_to_xfer;
  char outfilename[FILENAME_SIZE];
  wlong total=0;
  unsigned char header[CANONICAL_HEADER_SIZE];
  FILE *output;
  wave_info *joined_info;

  if (1 == join_output_dir_flag)
#ifdef HAVE_SNPRINTF
    snprintf(outfilename,FILENAME_SIZE,"%s/%s.%s",join_output_directory,PREFIX_JOIN,join_op->extension);
#else
    sprintf(outfilename,"%s/%s.%s",join_output_directory,PREFIX_JOIN,join_op->extension);
#endif
  else
#ifdef HAVE_SNPRINTF
    snprintf(outfilename,FILENAME_SIZE,"%s.%s",PREFIX_JOIN,join_op->extension);
#else
    sprintf(outfilename,"%s.%s",PREFIX_JOIN,join_op->extension);
#endif

  for (i=0;i<join_numfiles;i++)
    total += join_files[i]->data_size;

  if ((total % CD_BLOCK_SIZE) != 0) {
    join_pad_bytes = CD_BLOCK_SIZE - (total % CD_BLOCK_SIZE);
    total += join_pad_bytes;
  }

  if (1 == join_preview) {
    printf("\n");
    printf("Preview of changes:\n");
    printf("-------------------\n");
    printf("\n");
    printf("WAVE data from the following files:\n\n");
    for (i=0;i<join_numfiles;i++)
      printf("  %s\n",join_files[i]->filename);
    printf("\n");
    printf("will be joined into file '%s', in the order shown above.\n",outfilename);
    printf("\n");
    if (1 == join_pad) {
      if (0 != join_pad_bytes)
        printf("This file would then be padded with %d zero-bytes.\n",join_pad_bytes);
      else
        printf("This file wouldn't need any padding.\n");
    }
    else {
      if (0 != join_pad_bytes)
        printf("This file wouldn't be padded, but would need %d bytes of padding.\n",join_pad_bytes);
      else
        printf("This file wouldn't be padded, and wouldn't need it.\n");
    }
    return;
  }

  if (NULL == (joined_info = new_wave_info(NULL)))
    exit(1);

  if (NULL == (output = join_op->open_output(outfilename))) {
    error("could not open file '%s' for output",outfilename);
    exit(1);
  }

  joined_info->chunk_size = total + CANONICAL_HEADER_SIZE - 8;
  joined_info->channels = CD_CHANNELS;
  joined_info->samples_per_sec = CD_SAMPLES_PER_SEC;
  joined_info->avg_bytes_per_sec = CD_RATE;
  joined_info->block_align = CD_BLOCK_ALIGN;
  joined_info->bits_per_sample = CD_BITS_PER_SAMPLE;
  joined_info->data_size = total;
  joined_info->wave_format = WAVE_FORMAT_PCM;

  make_canonical_header(header,joined_info);

  if (write_n_bytes(output,header,CANONICAL_HEADER_SIZE) != CANONICAL_HEADER_SIZE) {
    error("error while writing %d-byte header",CANONICAL_HEADER_SIZE);
    exit(1);
  }

  for (i=0;i<join_numfiles;i++) {
    printf("Adding contents of %s to %s ... ",join_files[i]->filename,outfilename);
    fflush(stdout);
    if (NULL == (join_files[i]->input = join_files[i]->input_format->open_input(join_files[i]->filename))) {
      error("could not reopen file '%s' for input",join_files[i]->filename);
      exit(1);
    }
    bytes_to_skip = join_files[i]->header_size;
    while (bytes_to_skip > 0) {
      bytes_to_xfer = min(bytes_to_skip,CANONICAL_HEADER_SIZE);
      if (read_n_bytes(join_files[i]->input,header,bytes_to_xfer) != bytes_to_xfer) {
        error("error while reading %d bytes of data",bytes_to_xfer);
        exit(1);
      }
      bytes_to_skip -= bytes_to_xfer;
    }
    if (transfer_n_bytes(join_files[i]->input,output,join_files[i]->data_size) != join_files[i]->data_size) {
      error("error while transferring %lu bytes of data",join_files[i]->data_size);
      exit(1);
    }
    printf("done.\n");
    fclose(join_files[i]->input);
  }

  if (1 == join_pad) {
    if (0 != join_pad_bytes) {
      pad_file(output,join_pad_bytes);
      printf("Padded '%s' with %d zero-bytes.\n",outfilename,join_pad_bytes);
    }
    else
      printf("No padding needed for '%s'.\n",outfilename);
  }
  else {
    printf("File '%s' was not padded, ",outfilename);
    if (0 != join_pad_bytes)
      printf("though it needs %d bytes of padding.\n",join_pad_bytes);
    else
      printf("nor was it needed.\n");
  }

  fclose(output);
}

void process_files_join(int argc,char **argv,int start)
{
  int i,badfiles = 0,found_errors = 0;

  if (NULL == (join_files = malloc((join_numfiles + 1) * sizeof(wave_info *)))) {
    error("could not allocate memory for file info array");
    exit(1);
  }

  for (i=0;i<join_numfiles;i++)
    if (NULL == (join_files[i] = new_wave_info(argv[start+i])))
      badfiles++;

  join_files[join_numfiles] = NULL;

  if (badfiles) {
    error("could not open %d file%s, see above error messages",badfiles,(badfiles>1)?"s":"");
    exit(1);
  }

  /* validate files */
  for (i=0;i<join_numfiles;i++) {
    if (PROB_NOT_CD(join_files[i])) {
      error("file '%s' is not CD-quality",join_files[i]->filename);
      found_errors = 1;
    }
    if (PROB_HDR_INCONSISTENT(join_files[i])) {
      error("file '%s' has an inconsistent header",join_files[i]->filename);
      found_errors = 1;
    }
    if (PROB_TRUNCATED(join_files[i])) {
      error("file '%s' seems to be truncated",join_files[i]->filename);
      found_errors = 1;
    }
  }

  if (0 != found_errors) {
    error("could not join files due to errors, see above");
    exit(1);
  }

  if (1 == join_order)
    alter_file_order(join_files,join_numfiles);

  do_join();

  for (i=0;i<join_numfiles;i++)
    free(join_files[i]);

  free(join_files);
}

void join_main(int argc,char **argv,int start)
{
  int first_arg;

  parse_join(argc,argv,start,&first_arg);

  join_numfiles = argc-first_arg;

  process_files_join(argc,argv,first_arg);
}
