/*  mode_split.c - split mode module
 *  Copyright (C) 2000-2001  Jason Jordan (shnutils@freeshell.org)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "shntool.h"
#include "fileio.h"
#include "wave.h"
#include "misc.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define SPLIT_PREFIX "split-track"
#define MAX_PIECES   256

int split_point_file_flag = 0,
    split_prefix_flag = 0,
    split_output_dir_flag = 0,
    split_numfiles = 0,
    split_preview = 0,
    split_offset_flag = 0,
    split_offset = 1;

char *split_point_file = NULL,
     *split_prefix = SPLIT_PREFIX,
     *split_output_directory = ".";

wave_info *split_files[MAX_PIECES];

format_module *split_op = NULL;

void split_main(int,char **,int);

mode_module mode_split = {
  "split",
  "shnsplit",
  "splits CD-quality PCM WAVE data from one file into smaller files",
  split_main
};

void show_usage_split()
{
  int i, j = 0;

  printf("Usage: %s [OPTIONS] input_file\n",fullprogname);
  printf("\n");
  printf("Options:\n");
  printf("\n");
  printf("  -o fmt   specifies the output file format, where fmt is one of the following:\n");
  printf("\n");
  for (i=0;formats[i];i++)
    if (0 != formats[i]->supports_output)
      printf("             %" MAX_MODULE_NAME_LENGTH "s  (%s) %s\n",formats[i]->name,formats[i]->description,(0 == (j++))?"[DEFAULT]":"");
  printf("\n");
  printf("  -d dir   specifies alternate output directory (default is current directory)\n");
  printf("\n");
  printf("  -f file  specifies a file from which to read split point data\n");
  printf("\n");
  printf("  -n name  specifies a prefix for output files (default is \"%s\")\n",SPLIT_PREFIX);
  printf("\n");
  printf("  -c num   start counting from num when naming output files (default is 1)\n");
  printf("\n");
  printf("  -p       preview what changes would be made without actually making them\n");
  printf("\n");
  printf("  --       indicates that everything following it is a file name\n");
  printf("\n");
  printf("  -v       shows version information\n");
  printf("  -h       shows this help screen\n");
  printf("\n");
  printf("  If no split point file is given, then split points are read from standard\n");
  printf("  input.  See the man page (or README.txt) for more details about split points.\n");
  printf("\n");
  printf("  ------------\n");
  printf("  Sample uses:\n");
  printf("  ------------\n");
  printf("\n");
  printf("  %% %s -f offsets.txt -n bandYYYY-MM-DDd2t bigfile.wav\n",fullprogname);
  printf("\n");
  printf("  %% %s -p -d disc1 bigfile.shn < splitpoints\n",fullprogname);
  printf("\n");

  exit(1);
}

void parse_split(int argc,char **argv,int start,int *first_arg)
{
  int i,found_output = -1;

  for (i=0;formats[i];i++)
    if (0 != formats[i]->supports_output) {
      found_output = i;
      break;
    }

  if (-1 == found_output) {
    error("no output formats found, cannot continue");
    exit(1);
  }

  while (start < argc && argv[start][0] == '-') {
    if (argc > start && 0 == strcmp(argv[start],"--")) {
      start++;
      break;
    }
    else if (argc > start && 0 == strcmp(argv[start],"-v"))
      internal_version();
    else if (argc > start && 0 == strcmp(argv[start],"-h"))
      show_usage_split();
    else if (argc > start && 0 == strcmp(argv[start],"-p"))
      split_preview = 1;
    else if (argc > start && 0 == strcmp(argv[start],"-c")) {
      if (1 == split_offset_flag) {
        error("too many starting counts specified");
        type_h_for_help();
      }
      start++;
      split_offset_flag = 1;
      if (argc <= start) {
        error("missing starting count");
        type_h_for_help();
      }
      split_offset = atoi(argv[start]);
      if (split_offset < 0) {
        error("starting count cannot be negative");
        exit(1);
      }
    }
    else if (argc > start && 0 == strcmp(argv[start],"-d")) {
      if (1 == split_output_dir_flag) {
        error("too many output directories specified");
        type_h_for_help();
      }
      start++;
      split_output_dir_flag = 1;
      if (argc <= start) {
        error("missing output directory");
        type_h_for_help();
      }
      split_output_directory = argv[start];
    }
    else if (argc > start && 0 == strcmp(argv[start],"-n")) {
      if (1 == split_prefix_flag) {
        error("too many prefixes files specified");
        type_h_for_help();
      }
      start++;
      split_prefix_flag = 1;
      if (argc <= start) {
        error("missing prefix");
        type_h_for_help();
      }
      split_prefix = argv[start];
    }
    else if (argc > start && 0 == strcmp(argv[start],"-f")) {
      if (1 == split_point_file_flag) {
        error("too many split point files specified");
        type_h_for_help();
      }
      start++;
      split_point_file_flag = 1;
      if (argc <= start) {
        error("missing split point file");
        type_h_for_help();
      }
      split_point_file = argv[start];
    }
    else if (argc > start && 0 == strcmp(argv[start],"-o")) {
      if (NULL != split_op) {
        error("too many output file formats specified");
        type_h_for_help();
      }
      start++;
      if (argc <= start) {
        error("missing output file format");
        type_h_for_help();
      }
      for (i=0;formats[i];i++) {
        if (0 == strcmp(argv[start],formats[i]->name)) {
          if (0 == formats[i]->supports_output) {
            error("file format '%s' does not support output",formats[i]->name);
            type_h_for_help();
          }
          split_op = formats[i];
          break;
        }
      }
      if (NULL == split_op) {
        error("invalid output file format: %s",argv[start]);
        type_h_for_help();
      }
    }
    else {
      error("unknown argument: %s",argv[start]);
      type_h_for_help();
    }
    start++;
  }

  if (start != argc - 1) {
    error("need exactly one file to process");
    type_h_for_help();
  }

  if (NULL == split_op) {
    split_op = formats[found_output];
    printf("warning: no output format specified - defaulting to %s\n",split_op->name);
  }

  *first_arg = start;
}

void make_outfilename_split(char *outfilename,int num)
{
  if (1 == split_output_dir_flag)
#ifdef HAVE_SNPRINTF
    snprintf(outfilename,FILENAME_SIZE,"%s/%s%02d.%s",split_output_directory,split_prefix,num+split_offset,split_op->extension);
#else
    sprintf(outfilename,"%s/%s%02d.%s",split_output_directory,split_prefix,num+split_offset,split_op->extension);
#endif
  else
#ifdef HAVE_SNPRINTF
    snprintf(outfilename,FILENAME_SIZE,"%s%02d.%s",split_prefix,num+split_offset,split_op->extension);
#else
    sprintf(outfilename,"%s%02d.%s",split_prefix,num+split_offset,split_op->extension);
#endif
}

void split_file(wave_info *info)
{
  unsigned char header[CANONICAL_HEADER_SIZE];
  char outfilename[FILENAME_SIZE];
  int current;
  wint discard,bytes;
  FILE *output;

  if (NULL == (info->input = info->input_format->open_input(info->filename))) {
    error("could not reopen file '%s' for input",info->filename);
    exit(1);
  }

  discard = info->header_size;
  while (discard > 0) {
    bytes = min(discard,CANONICAL_HEADER_SIZE);
    if (read_n_bytes(info->input,header,bytes) != bytes) {
      error("error while discarding %d-byte header",info->header_size);
      exit(1);
    }
    discard -= bytes;
  }

  printf("\n");
  printf("Input file '%s' is being split into %d pieces.\n",info->filename,split_numfiles);
  printf("\n");

  for (current=0;current<split_numfiles;current++) {
    make_outfilename_split(outfilename,current);

    if (NULL == (output = split_op->open_output(outfilename))) {
      error("could not open file '%s' for output",outfilename);
      exit(1);
    }

    split_files[current]->chunk_size = split_files[current]->data_size + CANONICAL_HEADER_SIZE - 8;
    split_files[current]->channels = CD_CHANNELS;
    split_files[current]->samples_per_sec = CD_SAMPLES_PER_SEC;
    split_files[current]->avg_bytes_per_sec = CD_RATE;
    split_files[current]->block_align = CD_BLOCK_ALIGN;
    split_files[current]->bits_per_sample = CD_BITS_PER_SAMPLE;
    split_files[current]->wave_format = WAVE_FORMAT_PCM;
    split_files[current]->rate = CD_RATE;
    split_files[current]->length = split_files[current]->data_size / (wlong)CD_RATE;

    make_canonical_header(header,split_files[current]);

    length_to_str(split_files[current]);

    printf("Writing %s (%s) ... ",outfilename,split_files[current]->m_ss);

    fflush(stdout);

    if (write_n_bytes(output,header,CANONICAL_HEADER_SIZE) != CANONICAL_HEADER_SIZE) {
      error("error while writing %d-byte header",CANONICAL_HEADER_SIZE);
      exit(1);
    }

    if (transfer_n_bytes(info->input,output,split_files[current]->data_size) != split_files[current]->data_size) {
      error("error while transferring %ld bytes to output file",split_files[current]->data_size);
      exit(1);
    }

    printf("done.\n");

    fclose(output);
  }

  fclose(info->input);
}

void split_show_preview(char *filename)
{
  int i;
  char outfilename[FILENAME_SIZE];

  printf("\n");
  printf("Preview of changes:\n");
  printf("-------------------\n");
  printf("\n");
  printf("File '%s' will be split into %d pieces:\n",filename,split_numfiles);
  for (i=0;i<split_numfiles;i++) {
    make_outfilename_split(outfilename,i);
    printf("\n");

    split_files[i]->rate = CD_RATE;
    split_files[i]->length = split_files[i]->data_size / (wlong)CD_RATE;

    length_to_str(split_files[i]);

    printf("  + %s (%s)\n",outfilename,split_files[i]->m_ss);
    printf("    - this file will contain %ld bytes of WAVE data (plus %d-byte header)\n",split_files[i]->data_size,CANONICAL_HEADER_SIZE);
    if (PROB_TOO_SHORT(split_files[i]))
      printf("    - this file will be too short to be burned\n");
    if (PROB_BAD_BOUND(split_files[i]))
      printf("    - this file will not be cut on a sector boundary\n");
  }
}

int is_a_digit(char c)
{
  switch (c) {
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      break;
    default:
      return 0;
  }
  return 1;
}

int is_numeric(unsigned char *buf)
{
  unsigned char *p = buf;
  wlong bytes;

  while ('\0' != *p) {
    if (0 == is_a_digit(*p))
      return 0;
    p++;
  }

  bytes =
#ifdef HAVE_ATOL
      (wlong)atol(buf);
#else
      (wlong)atoi(buf);
#endif

  return bytes;
}

wlong is_m_ss(unsigned char *buf)
{
  unsigned char *colon,*p;
  int len,min,sec;
  wlong bytes;

  len = strlen(buf);

  if (len < 4)
    return 0;

  colon = buf + len - 3;

  if (':' != *colon)
    return 0;

  p = buf;
  while ('\0' != *p) {
    if (p != colon) {
      if (0 == is_a_digit(*p))
        return 0;
    }
    p++;
  }

  *colon = '\0';

  min = atoi(buf);
  sec = atoi(colon+1);

  if (sec >= 60) {
    error("invalid value for seconds: %d",sec);
    exit(1);
  }

  bytes = (wlong)(min * CD_RATE * 60) +
          (wlong)(sec * CD_RATE);

  return bytes;
}

wlong is_m_ss_ff(unsigned char *buf)
{
  unsigned char *colon,*dot,*p;
  int len,min,sec,frames;
  wlong bytes;

  len = strlen(buf);

  if (len < 7)
    return 0;

  colon = buf + len - 6;
  dot = buf + len - 3;

  if (':' != *colon || '.' != *dot)
    return 0;

  p = buf;
  while ('\0' != *p) {
    if (p != colon && p != dot) {
      if (0 == is_a_digit(*p))
        return 0;
    }
    p++;
  }

  *colon = '\0';
  *dot = '\0';

  min = atoi(buf);
  sec = atoi(colon+1);
  frames = atoi(dot+1);

  if (sec >= 60) {
    error("invalid value for seconds: %d",sec);
    exit(1);
  }

  if (frames >= 75) {
    error("invalid value for frames: %d",frames);
    exit(1);
  }

  bytes = (wlong)(min * CD_RATE * 60) +
          (wlong)(sec * CD_RATE) +
          (wlong)(frames * CD_BLOCK_SIZE);

  return bytes;
}

wlong is_m_ss_nnn(unsigned char *buf)
{
  unsigned char *colon,*dot,*p;
  int len,min,sec,ms,nearest_byte,nearest_frame;
  wlong bytes;

  len = strlen(buf);

  if (len < 8)
    return 0;

  colon = buf + len - 7;
  dot = buf + len - 4;

  if (':' != *colon || '.' != *dot)
    return 0;

  p = buf;
  while ('\0' != *p) {
    if (p != colon && p != dot) {
      if (0 == is_a_digit(*p))
        return 0;
    }
    p++;
  }

  *colon = '\0';
  *dot = '\0';

  min = atoi(buf);
  sec = atoi(colon+1);
  ms = atoi(dot+1);

  if (sec >= 60) {
    error("invalid value for seconds: %d",sec);
    exit(1);
  }

  nearest_byte = (int)((((double)ms * (double)CD_RATE) / 1000.0) + 0.5);
  nearest_frame = (int)((double)ms * .075 + 0.5) * CD_BLOCK_SIZE;

  bytes = (wlong)(min * CD_RATE * 60) +
          (wlong)(sec * CD_RATE);

  if (nearest_frame != nearest_byte)
    printf("warning: rounding %d:%02d.%03d (offset: %lu) to nearest sector boundary (offset: %lu)\n",
           min,sec,ms,bytes+(wlong)nearest_byte,bytes+(wlong)nearest_frame);

  bytes += (wlong)nearest_frame;

  return bytes;
}

wlong smrt_parse(unsigned char *data)
/* currently not so smart, so I gave it the Homer Simpson spelling :) */
{
  wlong bytes;

  while (data[strlen(data)-1] == '\n' ||
         data[strlen(data)-1] == '\t' ||
         data[strlen(data)-1] == ' ')
    data[strlen(data)-1] = '\0';

  /* check for all digits */
  if (0 != (bytes = is_numeric(data)))
    return bytes;

  /* check for m:ss */
  if (0 != (bytes = is_m_ss(data)))
    return bytes;

  /* check for m:ss.ff */
  if (0 != (bytes = is_m_ss_ff(data)))
    return bytes;

  /* check for m:ss.nnn */
  if (0 != (bytes = is_m_ss_nnn(data)))
    return bytes;

  /* it was not in any of these formats */

  error("split point '%s' is not in bytes, m:ss, m:ss.ff, or m:ss.nnn format",data);
  exit(1);

  return 0;
}

void read_split_points()
{
  FILE *split_fd;
  unsigned char data[BUF_SIZE];
  wlong previous = 0;

  if (1 == split_point_file_flag) {
    if (NULL == (split_fd = fopen(split_point_file,"rb"))) {
      error("could not open split point file '%s'",split_point_file);
      exit(1);
    }
  }
  else
    split_fd = stdin;

  fgets(data,BUF_SIZE-1,split_fd);

  while (0 == feof(split_fd)) {
    if (MAX_PIECES - 1 == split_numfiles) {
      error("too many split files would be created - maximum is %d",MAX_PIECES);
      exit(1);
    }

    if (NULL == (split_files[split_numfiles] = new_wave_info(NULL))) {
      error("could not allocate memory for split points array");
      exit(1);
    }

    split_files[split_numfiles]->beginning_byte = smrt_parse(data);
    if (split_files[split_numfiles]->beginning_byte <= previous) {
      error("split point %d comes before previous split point",split_numfiles+1);
      exit(1);
    }

    split_files[split_numfiles]->data_size = split_files[split_numfiles]->beginning_byte - previous;

    if (split_files[split_numfiles]->data_size < CD_MIN_BURNABLE_SIZE) {
      printf("warning: file %d will be too short to be burned\n",split_numfiles+1);
      split_files[split_numfiles]->problems |= PROBLEM_CD_BUT_TOO_SHORT;
      if (0 != split_files[split_numfiles]->data_size % CD_BLOCK_SIZE) {
        split_files[split_numfiles]->problems |= PROBLEM_CD_BUT_BAD_BOUND;
        printf("         and will not be cut on a sector boundary\n");
      }
    }
    else if (0 != split_files[split_numfiles]->data_size % CD_BLOCK_SIZE) {
      split_files[split_numfiles]->problems |= PROBLEM_CD_BUT_BAD_BOUND;
      printf("warning: file %d will not be cut on a sector boundary\n",split_numfiles+1);
    }

    previous = split_files[split_numfiles]->beginning_byte;
    split_numfiles++;
    fgets(data,BUF_SIZE-1,split_fd);
  }

  if (NULL == (split_files[split_numfiles] = new_wave_info(NULL))) {
    error("could not allocate memory for split points array");
    exit(1);
  }
  split_numfiles++;

  if (1 == split_point_file_flag)
    fclose(split_fd);

  if (1 == split_numfiles) {
    error("no split points given - nothing to do");
    exit(1);
  }
}

void process_file_split(char *filename)
{
  int i;
  wave_info *info;

  if (NULL == (info = new_wave_info(filename))) {
    error("cannot continue due to error(s) shown above");
    exit(1);
  }

  if (PROB_NOT_CD(info)) {
    error("cannot process %s because it is not CD-quality",info->filename);
    exit(1);
  }

  read_split_points();

  if (split_files[split_numfiles-2]->beginning_byte > info->data_size) {
    error("split points go beyond input file's data size");
    exit(1);
  }

  if (split_files[split_numfiles-2]->beginning_byte == info->data_size) {
    free(split_files[split_numfiles-1]);
    split_numfiles--;
  }
  else {
    split_files[split_numfiles-1]->beginning_byte = info->data_size;
    split_files[split_numfiles-1]->data_size = info->data_size - split_files[split_numfiles-2]->beginning_byte;

    if (split_files[split_numfiles-1]->data_size < CD_MIN_BURNABLE_SIZE) {
      printf("warning: file %d will be too short to be burned\n",split_numfiles);
      split_files[split_numfiles-1]->problems |= PROBLEM_CD_BUT_TOO_SHORT;
      if (0 != split_files[split_numfiles-1]->data_size % CD_BLOCK_SIZE) {
        split_files[split_numfiles-1]->problems |= PROBLEM_CD_BUT_BAD_BOUND;
        printf("         and will not be cut on a sector boundary\n");
      }
    }
    else if (0 != split_files[split_numfiles-1]->data_size % CD_BLOCK_SIZE) {
      split_files[split_numfiles-1]->problems |= PROBLEM_CD_BUT_BAD_BOUND;
      printf("warning: file %d will not be cut on a sector boundary\n",split_numfiles);
    }
  }

  if (1 == split_preview)
    split_show_preview(info->filename);
  else
    split_file(info);

  for (i=0;i<split_numfiles;i++)
    free(split_files[i]);

  free(info);
}

void process_split(int argc,char **argv,int start)
{
  char filename[FILENAME_SIZE];

  if (argc < start + 1) {
    /* no filenames were given, so we're reading one file from standard input. */
    fgets(filename,FILENAME_SIZE-1,stdin);
    if (0 == feof(stdin)) {
      filename[strlen(filename)-1] = '\0';
      process_file_split(filename);
    }
  }
  else
    process_file_split(argv[start]);
}

void split_main(int argc,char **argv,int start)
{
  int first_arg;

  parse_split(argc,argv,start,&first_arg);

  process_split(argc,argv,first_arg);
}
