/*  mode_strip.c - strip mode module
 *  Copyright (C) 2000-2001  Jason Jordan (shnutils@freeshell.org)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "shntool.h"
#include "convert.h"
#include "fileio.h"
#include "wave.h"
#include "misc.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define POSTFIX_STRIP "-stripped"

format_module *strip_op = NULL;

int strip_output_dir_flag = 0,
    strip_header = 1,
    strip_chunks = 1,
    strip_preview = 0;

char *strip_output_directory = ".";

void strip_main(int,char **,int);

mode_module mode_strip = {
  "strip",
  "shnstrip",
  "strips extra RIFF chunks and/or writes canonical headers",
  strip_main
};

void show_usage_strip()
{
  int i,j=0;

  printf("Usage: %s [OPTIONS] [files]\n",fullprogname);
  printf("\n");
  printf("Options:\n");
  printf("\n");
  printf("  -o fmt   specifies the output file format, where fmt is one of the following:\n");
  printf("\n");
  for (i=0;formats[i];i++)
    if (0 != formats[i]->supports_output)
      printf("             %" MAX_MODULE_NAME_LENGTH "s  (%s) %s\n",formats[i]->name,formats[i]->description,(0 == (j++))?"[DEFAULT]":"");
  printf("\n");
  printf("  -d dir   specifies alternate output directory (default is same as input file)\n");
  printf("\n");
  printf("  -nh      don't rewrite WAVE header in canonical format (default is yes)\n");
  printf("  -nc      don't strip unnecessary RIFF chunks (default is yes)\n");
  printf("\n");
  printf("  -p       preview what changes would be made without actually making them\n");
  printf("\n");
  printf("  --       indicates that everything following it is a file name\n");
  printf("\n");
  printf("  -v       shows version information\n");
  printf("  -h       shows this help screen\n");
  printf("\n");
  printf("  If no file names are given, then file names are read from standard input.\n");
  printf("\n");
  printf("  ------------\n");
  printf("  Sample uses:\n");
  printf("  ------------\n");
  printf("\n");
  printf("  %% %s -o shn -d /mnt/shn/stripped *.shn *.wav\n",fullprogname);
  printf("\n");
  printf("  %% find /your/audio/dir -type f | %s\n",fullprogname);
  printf("\n");

  exit(1);
}

void parse_strip(int argc,char **argv,int start,int *first_arg)
{
  int i,found_output = -1;

  for (i=0;formats[i];i++)
    if (0 != formats[i]->supports_output) {
      found_output = i;
      break;
    }

  if (-1 == found_output) {
    error("no output formats found, cannot continue");
    exit(1);
  }

  while (start < argc && argv[start][0] == '-') {
    if (argc > start && 0 == strcmp(argv[start],"--")) {
      start++;
      break;
    }
    else if (argc > start && 0 == strcmp(argv[start],"-v"))
      internal_version();
    else if (argc > start && 0 == strcmp(argv[start],"-h"))
      show_usage_strip();
    else if (argc > start && 0 == strcmp(argv[start],"-nh"))
      strip_header = 0;
    else if (argc > start && 0 == strcmp(argv[start],"-nc"))
      strip_chunks = 0;
    else if (argc > start && 0 == strcmp(argv[start],"-p"))
      strip_preview = 1;
    else if (argc > start && 0 == strcmp(argv[start],"-d")) {
      if (1 == strip_output_dir_flag) {
        error("too many output directories specified");
        type_h_for_help();
      }
      start++;
      strip_output_dir_flag = 1;
      if (argc <= start) {
        error("missing output directory");
        type_h_for_help();
      }
      strip_output_directory = argv[start];
    }
    else if (argc > start && 0 == strcmp(argv[start],"-o")) {
      if (NULL != strip_op) {
        error("too many output file formats specified");
        type_h_for_help();
      }
      start++;
      if (argc <= start) {
        error("missing output file format");
        type_h_for_help();
      }
      for (i=0;formats[i];i++) {
        if (0 == strcmp(argv[start],formats[i]->name)) {
          if (0 == formats[i]->supports_output) {
            error("file format '%s' does not support output",formats[i]->name);
            type_h_for_help();
          }
          strip_op = formats[i];
          break;
        }
      }
      if (NULL == strip_op) {
        error("invalid output file format: %s",argv[start]);
        type_h_for_help();
      }
    }
    else {
      error("unknown argument: %s",argv[start]);
      type_h_for_help();
    }
    start++;
  }

  if (0 == strip_header && 0 == strip_chunks) {
    error("nothing to strip if both -nh and -nc are specified");
    type_h_for_help();
  }

  if (NULL == strip_op) {
    strip_op = formats[found_output];
    printf("warning: no output format specified - defaulting to %s\n\n",strip_op->name);
  }

  *first_arg = start;
}

void make_outfilename_strip(char *infilename,char *outfilename)
{
  char *basename;

  if ((basename = strrchr(infilename,'/')))
    basename++;
  else
    basename = infilename;

  if (1 == strip_output_dir_flag)
#ifdef HAVE_SNPRINTF
    snprintf(outfilename,FILENAME_SIZE,"%s/%s",strip_output_directory,basename);
#else
    sprintf(outfilename,"%s/%s",strip_output_directory,basename);
#endif
  else
#ifdef HAVE_SNPRINTF
    snprintf(outfilename,FILENAME_SIZE,"%s",infilename);
#else
    sprintf(outfilename,"%s",infilename);
#endif

  if (filename_contains_a_dot(outfilename))
    *(strrchr(outfilename,'.')) = '\0';

  strcat(outfilename,POSTFIX_STRIP);
  strcat(outfilename,".");
  strcat(outfilename,strip_op->extension);
}

void strip_and_canonicize(wave_info *info)
{
  wint new_header_size;
  wlong new_chunk_size,possible_extra_stuff;
  unsigned char *header;
  char outfilename[FILENAME_SIZE];
  FILE *output;

  if (PROB_HDR_INCONSISTENT(info)) {
    error("skipping file '%s' because its header is inconsistent",info->filename);
    return;
  }

  if (PROB_TRUNCATED(info)) {
    error("skipping file '%s' because its seems to be truncated",info->filename);
    return;
  }

  if ((1 == strip_header && 1 == strip_chunks) && ((0 == PROB_EXTRA_CHUNKS(info)) && (0 == PROB_HDR_NOT_CANONICAL(info)))) {
    error("skipping file '%s' because it already has a canonical header and no extra RIFF chunks",info->filename);
    return;
  }

  if ((1 == strip_chunks && 0 == strip_header) && (0 == PROB_EXTRA_CHUNKS(info))) {
    error("skipping file '%s' because it already has no extra RIFF chunks",info->filename);
    return;
  }

  if ((1 == strip_header && 0 == strip_chunks) && (0 == PROB_HDR_NOT_CANONICAL(info))) {
    error("skipping file '%s' because it already has a canonical header",info->filename);
    return;
  }

  new_header_size = info->header_size;
  new_chunk_size = info->chunk_size - (info->header_size - CANONICAL_HEADER_SIZE);
  possible_extra_stuff = info->total_size - (info->data_size + info->header_size);

  make_outfilename_strip(info->filename,outfilename);

  if (1 == strip_preview) {
    printf("%s --> %s\n",info->filename,outfilename);
    if (1 == strip_header && PROB_HDR_NOT_CANONICAL(info))
      printf("  - will rewrite %d-byte WAVE header to the canonical %d-byte header\n",info->header_size,CANONICAL_HEADER_SIZE);
    if (1 == strip_chunks && PROB_EXTRA_CHUNKS(info))
      printf("  - will strip %lu byte%s worth of extra RIFF chunk(s) from the end of this file\n",possible_extra_stuff,(possible_extra_stuff > 1)?"s":"");
    printf("\n");
    return;
  }

  if (1 == strip_header)
    new_header_size = CANONICAL_HEADER_SIZE;

  if (1 == strip_chunks) {
    new_chunk_size = (long)new_header_size + info->data_size - 8;
    possible_extra_stuff = 0;
  }
  info->new_data_size = info->data_size + possible_extra_stuff;

  if (NULL == (info->input = info->input_format->open_input(info->filename))) {
    error("could not reopen file '%s' for input",info->filename);
    return;
  }

  if (NULL == (output = strip_op->open_output(outfilename))) {
    error("could not open file '%s' for output",outfilename);
    return;
  }

  if (NULL == (header = malloc(info->header_size * sizeof(unsigned char)))) {
    error("could not allocate %d bytes for WAVE header",info->header_size);
    return;
  }

  if (read_n_bytes(info->input,header,info->header_size) != info->header_size) {
    error("error while reading %d bytes of data, skipping this file",info->header_size);
    return;
  }

  /* already read in header, now check to see if we need to overwrite that with a canonical one */
  if (1 == strip_header)
    make_canonical_header(header,info);

  ulong_to_uchar_le(header+4,new_chunk_size);

  if (write_n_bytes(output,header,new_header_size) != new_header_size) {
    error("error while writing %d bytes of data, skipping this file",new_header_size);
    return;
  }

  printf("%s --> %s ... ",info->filename,outfilename);
  fflush(stdout);

  if (transfer_n_bytes(info->input,output,info->new_data_size) != info->new_data_size) {
    error("error while writing %lu bytes of data, skipping this file",info->new_data_size);
    return;
  }

  printf("done.\n");

  free(header);

  fclose(info->input);
  fclose(output);
}

void process_file_strip(char *filename)
{
  wave_info *info;

  if (NULL == (info = new_wave_info(filename)))
    return;

  strip_and_canonicize(info);

  free(info);
}

void process_strip(int argc,char **argv,int start)
{
  int i;
  char filename[FILENAME_SIZE];

  if (1 == strip_preview) {
    printf("\n");
    printf("Preview of changes:\n");
    printf("-------------------\n");
    printf("\n");
  }
  if (argc < start + 1) {
    /* no filenames were given, so we're reading files from standard input.  */
    fgets(filename,FILENAME_SIZE-1,stdin);
    while (0 == feof(stdin)) {
      filename[strlen(filename)-1] = '\0';
      process_file_strip(filename);
      fgets(filename,FILENAME_SIZE-1,stdin);
    }
  }
  else
    for (i=start;i<argc;i++)
      process_file_strip(argv[i]);
}

void strip_main(int argc,char **argv,int start)
{
  int first_arg;

  parse_strip(argc,argv,start,&first_arg);

  process_strip(argc,argv,first_arg);
}
